{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "44056652",
   "metadata": {},
   "source": [
    "# Week VI - Radioactive Decay, 1D Random Walks and Duck Typing"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9b494a8e",
   "metadata": {},
   "source": [
    "## Simulating Radioactive Decay\n",
    "\n",
    "In the following Python class we implement a simulator of radioactive decay for N isotopes.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "d8f25380",
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "import matplotlib.pyplot as plt\n",
    "\n",
    "class radio_decay:\n",
    "    '''\n",
    "    This class simulates the radioactive decay\n",
    "    of N isotopes\n",
    "    '''\n",
    "    \n",
    "    def __init__(self,N_init,lamb):\n",
    "        'Initialise the process'\n",
    "        self.N_init = N_init\n",
    "        self.lamb = lamb\n",
    "        self.N = N_init\n",
    "        self.rng = np.random.default_rng(seed=42424)\n",
    "        self.step_counter = 0\n",
    "        self.all_N = [N_init]\n",
    "        print('Initial number of particles: {}, decay constant: {}'.format(self.N_init,self.lamb))\n",
    "    \n",
    "    def run_single_step(self):\n",
    "        'Run a single random step for N particles'\n",
    "        r = self.rng.random(self.N)\n",
    "        self.N = self.N - sum(map(lambda x: True if (x <= self.lamb) else False, r))\n",
    "        \n",
    "    def run_steps(self,n_steps):\n",
    "        'Run a number of random steps equal to n_steps'\n",
    "        for i in range(n_steps):\n",
    "            self.step_counter += 1\n",
    "            print('Running step # {}'.format(self.step_counter))\n",
    "            self.run_single_step()\n",
    "            print('Number of particles: {}'.format(self.N))\n",
    "            self.all_N.append(self.N)\n",
    "    \n",
    "    def visualize(self,logplot=False):\n",
    "        if logplot:\n",
    "            plt.semilogy(range(len(self.all_N)),self.all_N,'o-')\n",
    "        else:\n",
    "            plt.plot(range(len(self.all_N)),self.all_N,'o-')\n",
    "        plt.xlabel('t')\n",
    "        plt.ylabel('N')\n",
    "        plt.show()\n",
    "        \n",
    "    def status(self):\n",
    "        print('Current number of particles: {}, after {} steps'.format(self.N,self.step_counter))\n",
    "        "
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3a075598",
   "metadata": {},
   "outputs": [],
   "source": [
    "#Let's create an instance of the class\n",
    "rd = radio_decay(10000,0.1)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "35ec1ca3",
   "metadata": {},
   "outputs": [],
   "source": [
    "#Let's create an instance of the class\n",
    "rd = radio_decay(10000,0.1)\n",
    "#and evolve it for ten steps\n",
    "rd.run_steps(10)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "0177033b",
   "metadata": {},
   "outputs": [],
   "source": [
    "#We inspect the status\n",
    "rd.status()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "396cca04",
   "metadata": {},
   "outputs": [],
   "source": [
    "#Here we visualize data\n",
    "rd.visualize()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e22f8afd",
   "metadata": {},
   "outputs": [],
   "source": [
    "#and evolve it for 50 steps\n",
    "rd.run_steps(50)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "657e911e",
   "metadata": {},
   "outputs": [],
   "source": [
    "rd.visualize()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "54a77635",
   "metadata": {},
   "outputs": [],
   "source": [
    "#Homework: try to fit the exponential and extract lambda\n",
    "import matplotlib.pyplot as plt\n",
    "plt.semilogy(range(len(rd.all_N)),rd.all_N,'o-')"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e0f34ed8",
   "metadata": {},
   "outputs": [],
   "source": []
  },
  {
   "cell_type": "markdown",
   "id": "ac7de660",
   "metadata": {},
   "source": [
    "-------------------------------------------------------"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "39b34938",
   "metadata": {},
   "source": [
    "## 1D Random Walks\n",
    "\n",
    "Next, we implement a Python class to simulate an ensemble of random walkers in one dimension"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "384f2075",
   "metadata": {},
   "outputs": [],
   "source": [
    "import sys\n",
    "import numpy as np \n",
    "\n",
    "class rw_1d:\n",
    "    '''\n",
    "    This class simulates the evolution\n",
    "    of an ensamble of random walkers\n",
    "    '''\n",
    "    def __init__(self,N_steps,N_walkers,seeds = [42424],len_step=1):\n",
    "        'Initialise the process'\n",
    "        self.N_steps = N_steps\n",
    "        self.N_walkers = N_walkers\n",
    "        self.len_step = len_step\n",
    "        if seeds == 'auto':\n",
    "            seeds = range(self.N_walkers)\n",
    "        else:\n",
    "            if len(seeds)!=N_walkers:\n",
    "                sys.exit('ERROR: seed length is not equal to number of random walkers')\n",
    "        self.seeds = seeds\n",
    "        self.rngs = [np.random.default_rng(seed=i) for i in self.seeds]\n",
    "        print('Number of steps: {}, number of walkers: {}'.format(self.N_steps,self.N_walkers))\n",
    "        self.x_trajs = []\n",
    "        self.x_av = 0\n",
    "        self.x2_av = 0\n",
    "        self.var = 0\n",
    "    \n",
    "    def run_rw(self,rw_index):\n",
    "        '''\n",
    "        Evolve one random walker\n",
    "        '''\n",
    "        r = self.rngs[rw_index].random(self.N_steps)\n",
    "        steps = list(map(lambda x: -1 if x<0.5 else 1, r))\n",
    "        steps = np.array(steps)\n",
    "        x = np.sum(steps)\n",
    "        x2 = np.sum(steps)**2\n",
    "        return x,x2,np.cumsum(steps)\n",
    "    \n",
    "    def run_many_rw(self):\n",
    "        '''\n",
    "        Evolve many random walkers\n",
    "        '''\n",
    "        x_l = []\n",
    "        x2_l = []\n",
    "        for i in range(self.N_walkers):\n",
    "            x,x2,traj= self.run_rw(i)\n",
    "            self.x_trajs.append(traj) \n",
    "            x_l.append(x)\n",
    "            x2_l.append(x2)\n",
    "        self.x_av = np.average(x_l)\n",
    "        self.x2_av = np.average(x2_l)\n",
    "        self.var  = self.x2_av - self.x_av**2\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "b8157095",
   "metadata": {},
   "outputs": [],
   "source": [
    "r = rw_1d(N_steps=1000,N_walkers=100,seeds='auto')\n",
    "r.run_many_rw()\n",
    "print(r.x_av,r.x2_av,r.var)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "953bfb05",
   "metadata": {},
   "outputs": [],
   "source": [
    "import matplotlib.pyplot as plt\n",
    "plt.xlabel('steps')\n",
    "plt.ylabel('$x$')\n",
    "_ = [plt.plot(np.arange(r.N_steps),r.x_trajs[i],label='walker {}'.format(i+1)) for i in range(r.N_walkers)]\n",
    "plt.legend()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "ee52bdf8",
   "metadata": {},
   "outputs": [],
   "source": [
    "import matplotlib.pyplot as plt\n",
    "a = np.array(r.x_trajs)**2\n",
    "a = np.average(a,axis=0)\n",
    "plt.xlabel('steps')\n",
    "plt.ylabel('$x^2$')\n",
    "_ = [plt.plot(np.arange(r.N_steps),r.x_trajs[i]**2) for i in range(r.N_walkers)]\n",
    "plt.plot(np.arange(r.N_steps),a,'-',lw=2,c='r',label='average over walkers')\n",
    "plt.plot(np.arange(r.N_steps),np.arange(r.N_steps),'-',lw=2,c='k',label='theory')\n",
    "plt.legend()"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e98a034c",
   "metadata": {},
   "source": [
    "-------------------------------------------------"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e132b389",
   "metadata": {},
   "source": [
    "## Duck Typing\n",
    "\n",
    "> If it walks like a duck and it quacks like a duck, then it must be a duck.\n",
    "\n",
    "Often, we do not really care about the type of an object but we are interested only in its behaviour. In particular, we typically need to know whether certain methods are implemented.\n",
    "\n",
    "**Duck typing** is a form of polymorphims where functions operate on any object that implements the appropiate methods, regardless of their classes or explicit interface declarations (L. Ramalho, *Fluent Python*, O'Reilly, 2016).\n",
    "\n",
    "Let us do some tests:"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5b250f35",
   "metadata": {},
   "outputs": [],
   "source": [
    "iter??"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "e5dceb9c",
   "metadata": {},
   "outputs": [],
   "source": [
    "import numpy as np\n",
    "def check_iterable(x):\n",
    "    try:\n",
    "        iter(x)\n",
    "        print('{} is iterable'.format(x))\n",
    "        return True\n",
    "    except TypeError as err:\n",
    "        print(err)\n",
    "        return False\n",
    "for x in [np.array([1,2,3]),[1,2,3],(1,2,3),'abcd',2]:\n",
    "    check_iterable(x)\n",
    "    "
   ]
  },
  {
   "cell_type": "markdown",
   "id": "35e60bf4",
   "metadata": {},
   "source": [
    "### Generators and the iterator protocol\n",
    "\n",
    "In Python, there is a consistent way to iterate over sequences, no matter the specific object type. This happens thanks to an **iterator protocol**, which is a generic way to make objects iterable.\n",
    "**Generators** allow to construct iterable objects, in a rather concise manner.\n"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "90ca30be",
   "metadata": {},
   "outputs": [],
   "source": [
    "range(10)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "c723b10a",
   "metadata": {},
   "outputs": [],
   "source": [
    "np.arange(10)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "61373610",
   "metadata": {},
   "outputs": [],
   "source": [
    "list(range(10))"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "4fe01811",
   "metadata": {},
   "outputs": [],
   "source": [
    "def gen_cube(N):\n",
    "    for i in range(N):\n",
    "        yield (i+1)**3\n",
    "gen_cube(8)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "277f03ae",
   "metadata": {},
   "outputs": [],
   "source": [
    "gen = gen_cube(8)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "5f56c36a",
   "metadata": {},
   "outputs": [],
   "source": [
    "for i in gen:\n",
    "    print(i)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "3caa40aa",
   "metadata": {},
   "outputs": [],
   "source": [
    "for i in map(lambda x: x**2, [1,2,3]):\n",
    "    print(i+2)"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9c187cb6",
   "metadata": {},
   "outputs": [],
   "source": [
    "map(lambda x: x**2, [1,2,3])"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "id": "9aa515d5",
   "metadata": {},
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "fis_comp_2023",
   "language": "python",
   "name": "fis_comp_2023"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.11.1"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
